;; Generated package description from mines.el  -*- no-byte-compile: t -*-
(define-package "mines" "1.6.0.20201130.184335" "Minesweeper game" '((emacs "24.4") (cl-lib "0.5")) :commit "97e0a054523de2a313d9c9e678f77985b284ca3c" :authors '(("Tino Calancha" . "tino.calancha@gmail.com")) :maintainer '("Tino Calancha" . "tino.calancha@gmail.com") :keywords '("games") :url "https://github.com/calancha/Minesweeper")
